/******************************************************************************/
/*                                                                            */
/* Module     : gvibConnection.i                                              */
/*                                                                            */
/* Description: This module is the SWIG file containing all of the C code     */
/*              associated withe python Connection class and which is used to */
/*              interface with the Interbase API library.                     */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Copyright (c) 2000 Going Virtual, LLC. All rights reserved.                */
/*                                                                            */
/* This library is free software; you can redistribute it and/or              */
/* modify it under the terms of the GNU Lesser General Public                 */
/* License as published by the Free Software Foundation; either               */
/* version 2.1 of the License, or (at your option) any later version.         */
/*                                                                            */
/* This library is distributed in the hope that it will be useful,            */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of             */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          */
/* Lesser General Public License for more details.                            */
/*                                                                            */
/* You should have received a copy of the GNU Lesser General Public           */
/* License along with this library; if not, write to the Free Software        */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */
/*                                                                            */
/*                                                                            */
/* Contact Information:                                                       */
/*                                                                            */
/* Going Virtual, LLC                                                         */
/* 10800 Independence Pointe Pkwy                                             */
/* Matthews, NC 28105                                                         */
/* www.goingv.com                                                             */
/*                                                                            */
/* Robert E. Tierney                                                          */
/* bob.tierney@goingv.com                                                     */
/* 1 (704) 849-0731 x109                                                      */
/*                                                                            */
/******************************************************************************/
/*
 * $Id: gvibConnect.i,v 1.2 2000/07/11 20:03:59 btierney Exp $
 * $Revision: 1.2 $
 */

%{  /* Start of SWIG-Free Section

/*----------------------------------------------------------------------------*/
/* Structure for use in creating a 'ConnectionType' python object type.       */
/*----------------------------------------------------------------------------*/
static PyTypeObject ConnectionType =
{
  PyObject_HEAD_INIT(&PyType_Type)                          /* initialization */
  0,                                                        /* ob_size        */
  "Connection",                                             /* tp_name        */
  sizeof(Connection),                                       /* tp_basicsize   */
  0,                                                        /* tp_itemsize    */

  (destructor)          IBConnectDeallocate,
  (printfunc)           0,
  (getattrfunc)         IBConnectGetAttr,
  (setattrfunc)         0,
  (cmpfunc)             0,
  (reprfunc)            0,
  (PyNumberMethods *)   0,
  (PySequenceMethods *) 0,
  (PyMappingMethods *)  0,
  (hashfunc)            0,
  (ternaryfunc)         0,
  (reprfunc)            0,
  (getattrofunc)        0,
  (setattrofunc)        0,
  (PyBufferProcs *)     0,

  0,                                                        /* tp_flags       */
  "Interbase Database Connection"                           /* tp_doc         */
};

%}  /* End of SWIG-Free Section */

%inline
%{

/******************************************************************************/
/*                                                                            */
/* IBConnectGetAttr                                                           */
/*                                                                            */
/* This function returns data from the Connection object to python in a form  */
/* that can then be passed from python to an Interbase API function.          */
/*                                                                            */
/******************************************************************************/
static PyObject *IBConnectGetAttr (Connection *self, char *name)
{
  char tmp_string[256];

  if (strcmp(name, "status") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) &self->status, "_ISC_STATUS_p");
    return PyString_FromString(tmp_string);
  }

  if (strcmp(name, "db_handle") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) &self->db_handle, "_isc_db_handle_p");
    return PyString_FromString(tmp_string);
  }

  if (strcmp(name, "tr_handle") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) &self->tr_handle, "_isc_tr_handle_p");
    return PyString_FromString(tmp_string);
  }
}

/******************************************************************************/
/*                                                                            */
/* IBConnect                                                                  */
/*                                                                            */
/* This function makes a connection to an Interbase database.  It is called   */
/* from python.  Once connected a Transaction is started to 1) turn off       */
/* auto-commit and 2) eliminate the need for the user to manage transactions. */
/* When a user issues a commit from python, the transaction is committed but  */
/* the transaction context is retained so it can be reused for further SQL    */
/* work by the user.                                                          */
/*                                                                            */
/******************************************************************************/
static PyObject *
IBConnect (char *db_name, char *user, char *pwd, PyObject *tpb_parms)
{
  Connection *self;
  long  rc;

  char *dpb;
  char  dpbBuffer[256];
  long  dpbLength;

  char *tpb = NULL;
  char  tpbBuffer[128];
  long  tpbLength;

  short length;

  /*--------------------------------------------------------------------------*/
  /* Allocate a Connection object. PyObject_NEW sets the error conndition so  */
  /* we just need to return NULL.                                             */
  /*--------------------------------------------------------------------------*/
  self = PyObject_NEW(Connection, &ConnectionType);
  if (!self)
    return NULL;

  self->db_handle = 0L;
  self->tr_handle = 0L;
  self->page_size = 1024;
  memset((char *) self->status, 0, sizeof(self->status));

  /*--------------------------------------------------------------------------*/
  /* Initialize the database parameter block.  We do it manually instead of   */
  /* using isc_expand_dpb() because I couldn't get that darn thing to work    */
  /* correctly.  :-(                                                          */
  /*--------------------------------------------------------------------------*/
  dpb    = dpbBuffer;
  *dpb++ = isc_dpb_version1;

  *dpb++ = isc_dpb_user_name;
  *dpb++ = length = strlen(user);
  memcpy(dpb, user, length);
  dpb += length;

  *dpb++ = isc_dpb_password;
  *dpb++ = length = strlen(pwd);
  memcpy(dpb, pwd, length);
  dpb += length;

  dpbLength = dpb - dpbBuffer;

  /*--------------------------------------------------------------------------*/
  /* Initialize the transaction parameter buffer.                             */
  /*--------------------------------------------------------------------------*/
  tpbLength = PyTuple_Size(tpb_parms);
  if (tpbLength > 0)
  {
    PyObject *tmp;

    tpb = tpbBuffer;
    for (rc = 0; rc < tpbLength; rc++)
    {
      if (tmp = PyTuple_GetItem(tpb_parms, rc))
      {
        tpbBuffer[rc] = (char) PyInt_AsLong(tmp);
      }
    }
  }

  /*--------------------------------------------------------------------------*/
  /* Make the database connection and start a transaction.                    */
  /*--------------------------------------------------------------------------*/
  Py_BEGIN_ALLOW_THREADS

  rc = isc_attach_database(self->status,
  		strlen(db_name),
		db_name,
		&self->db_handle,
		dpbLength,
		dpbBuffer);

  if (rc == 0)
  {
    rc = isc_start_transaction(self->status,
                &self->tr_handle,
		1,
		&self->db_handle,
		(unsigned short) tpbLength,
		tpb);
  }

  if (rc == 0)
  {
    char items[] = { isc_info_page_size, isc_info_end };

    rc = isc_database_info(self->status,
                &self->db_handle,
		sizeof(items),
		items,
		sizeof(dpbBuffer),
		dpbBuffer);

    if (*dpbBuffer == isc_info_page_size)
    {
      short length = isc_vax_integer(dpbBuffer + 1, 2);
      self->page_size = isc_vax_integer(dpbBuffer + 3, length);
    }
  }
   
  Py_END_ALLOW_THREADS

  /*--------------------------------------------------------------------------*/
  /* If there was an error set the exception object and return NULL.          */
  /*--------------------------------------------------------------------------*/
  if (DB_ERROR(self->status))
  {
    (void) IBSetException(self->status);

    Py_DECREF(self);

    return NULL;
  }

  return (PyObject *) self;
}



/******************************************************************************/
/*                                                                            */
/* IBConnectDeallocate                                                        */
/*                                                                            */
/* This function detaches from a database connection and frees the memory     */
/* associated with the connection object.                                     */
/*                                                                            */
/******************************************************************************/
static void
IBConnectDeallocate (Connection *self)
{
  if (self)
  {
    Py_BEGIN_ALLOW_THREADS
      
    if (self->tr_handle)
      isc_rollback_transaction(self->status, &self->tr_handle);

    if (self->db_handle)
      isc_detach_database(self->status, &self->db_handle);

    Py_END_ALLOW_THREADS
  }

  PyMem_DEL(self);
}

%}
