/*                                                                            */
/* Module     : gvibCursor.i                                                  */
/*                                                                            */
/* Description: This module is the SWIG source file containing all of the     */
/*              C code associated with the python Cursor class and which      */
/*              is used to interface with the Interbase API library.  This    */
/*              code implements utility-like functions that mainly deal with  */
/*              binding input and ouput columns.  A 'Cursor' object is        */
/*              implemented herein to contain all of the data structures      */
/*              needed by Interbase to execute SQL statements.                */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Copyright (c) 2000 Going Virtual, LLC. All rights reserved.                */
/*                                                                            */
/* This library is free software; you can redistribute it and/or              */
/* modify it under the terms of the GNU Lesser General Public                 */
/* License as published by the Free Software Foundation; either               */
/* version 2.1 of the License, or (at your option) any later version.         */
/*                                                                            */
/* This library is distributed in the hope that it will be useful,            */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of             */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          */
/* Lesser General Public License for more details.                            */
/*                                                                            */
/* You should have received a copy of the GNU Lesser General Public           */
/* License along with this library; if not, write to the Free Software        */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */
/*                                                                            */
/*                                                                            */
/* Contact Information:                                                       */
/*                                                                            */
/* Going Virtual, LLC                                                         */
/* 10800 Independence Pointe Pkwy                                             */
/* Matthews, NC 28105                                                         */
/* www.goingv.com                                                             */
/*                                                                            */
/* Robert E. Tierney                                                          */
/* bob.tierney@goingv.com                                                     */
/* 1 (704) 849-0731 x109                                                      */
/*                                                                            */
/******************************************************************************/
/* Change History                                                             */
/*                                                                            */
/* 07/31/2000 - Fixed bugs in IBCursorFreeCols and IBCursorReallocXDA. Thanks */
/*              to Brad Clements.                                             */
/*                                                                            */
/******************************************************************************/
/*
 * $Id: gvibCursor.i,v 1.7 2000/07/31 13:41:05 btierney Exp $
 * $Revision: 1.7 $
 */

%{  /* Start of SWIG-Free Section */

/*----------------------------------------------------------------------------*/
/* The C extension CursorType object table.                                   */
/*----------------------------------------------------------------------------*/
static PyTypeObject CursorType =
{
  PyObject_HEAD_INIT(&PyType_Type)                          /* initialization */
  0,                                                        /* ob_size        */
  "Cursor",                                                 /* tp_name        */
  sizeof(Cursor),                                           /* tp_basicsize   */
  0,                                                        /* tp_itemsize    */

  (destructor)          IBCursorDeallocate,
  (printfunc)           0,
  (getattrfunc)         IBCursorGetAttr,
  (setattrfunc)         IBCursorSetAttr,
  (cmpfunc)             0,
  (reprfunc)            0,
  (PyNumberMethods *)   0,
  (PySequenceMethods *) 0,
  (PyMappingMethods *)  0,
  (hashfunc)            0,
  (ternaryfunc)         0,
  (reprfunc)            0,
  (getattrofunc)        0,
  (setattrofunc)        0,
  (PyBufferProcs *)     0,

  0,                                                        /* tp_flags       */
  "Interbase Database Cursor"                               /* tp_doc         */
};

%}  /* End of SWIG-Free Section */


%inline
%{
/******************************************************************************/
/*                                                                            */
/* IBCursorGetColDesc                                                         */
/*                                                                            */
/* This function loops through the XSQLVAR array of the output XSQLDA and     */
/* constructs a list of tuples.  Each tuple represents one column in the      */
/* the result set.                                                            */
/*                                                                            */
/******************************************************************************/
static PyObject *
IBCursorGetColDesc (Cursor *self)
{
  short i;
  char buffer[512];
  XSQLDA  *sqlda;
  XSQLVAR *var;

  PyObject *PyMsg;
  PyObject *PyList;
  PyObject *PyTuple;

  char  alias[64];
  char  column[64];
  short type;
  short display_size = 0;
  short internal_size;
  short precision;
  short scale;
  short null_ok;
  short subtype;

  sqlda = self->osqlda;
  var   = self->osqlda->sqlvar;

  if ((PyList = PyList_New(sqlda->sqld)) == NULL)
    return NULL;

  for (i = 0; i < sqlda->sqld; i++, var++)
  {
    memset(column, 0, sizeof(column));
    memcpy(column, var->sqlname, var->sqlname_length);

    memset(alias, 0, sizeof(alias));
    memcpy(alias, var->aliasname, var->aliasname_length);

    type          = var->sqltype & ~1;
    internal_size = var->sqllen;
    scale         = var->sqlscale * -1;
    null_ok       = var->sqltype & 1;
    subtype       = var->sqlsubtype;

    /*------------------------------------------------------------------------*/
    /* Unfortunately Interbase does not store the precision value in the      */
    /* XSQLVAR data structure.  Consequently we have to make a database call  */
    /* to the system tables to get the precision.                             */
    /*------------------------------------------------------------------------*/
    precision = getPrecision(self, column);
    if (precision < 0)
    {
      Py_DECREF(PyList);
      return NULL;
    }

    PyTuple = Py_BuildValue("(sshhhhhhh)", 
                          column, 
			  alias,
			  type, 
			  display_size, 
			  internal_size, 
			  precision,
			  scale,
			  null_ok,
			  subtype);

    if (PyList_SetItem(PyList, i, PyTuple))
    {
      Py_DECREF(PyList);
      return NULL;
    }
  }

  return PyList;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorFreeCols                                                           */
/*                                                                            */
/* This function loops through the XSQLVAR structures and frees the memory    */
/* allocated to the data area and the null indicator.  This method is called  */
/* from the IBCursorReallocXDA function.  This is the only case where it      */
/* should be called.                                                          */
/*                                                                            */
/******************************************************************************/
static long IBCursorFreeCols (Cursor *self, char *type)
{
  short i, n;
  XSQLDA  *sqlda;
  XSQLVAR *var;

  if (strcmp(type, "isqlda") == 0)
  {
    sqlda = self->isqlda;
    var   = self->isqlda->sqlvar;
  }
  else if (strcmp(type, "osqlda") == 0)
  {
    sqlda = self->osqlda;
    var   = self->osqlda->sqlvar;
  }
  else
  {
    return -1;
  }

  /*--------------------------------------------------------------------------*/
  /* Change sqld to sqln. Found by Brad Clements. 07/31/2000                  */
  /*--------------------------------------------------------------------------*/
  for (i = 0, n = sqlda->sqln; i < n; i++, var++) 
  {
    Py_Free(var->sqldata);
    Py_Free(var->sqlind); 
  }

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorGetRow                                                             */
/*                                                                            */
/* Retrieve a single row from the XSQLVAR data structure.  This function      */
/* assumes that a fetch has already been done.  Data is returned to the       */
/* caller as a tuple object.                                                  */
/*                                                                            */
/******************************************************************************/
static PyObject *IBCursorGetRow (Cursor *self)
{
  char *ptr;
  char  buffer[128];
  char  table[33];
  char  column[33];
  int   ctype;
  short i, n, k;
  short length;
  long  divisor;

  struct tm dtime;

  XSQLDA  *sqlda;
  XSQLVAR *ovar;

  PyObject *result = NULL;
  PyObject *PyTmp  = NULL;

  sqlda = self->osqlda;
  ovar  = self->osqlda->sqlvar;

  if ((result = PyTuple_New(sqlda->sqld)) == NULL)
    return NULL;

  /*--------------------------------------------------------------------------*/
  /* Loop through the XSQLVAR array and pull the data from each (column).     */
  /*--------------------------------------------------------------------------*/
  for (i = 0, n = sqlda->sqld; i < n; i++, ovar++)
  {
    /*------------------------------------------------------------------------*/
    /* If the column contains null then return the python None object.        */
    /*------------------------------------------------------------------------*/
    if ((ovar->sqltype & 1) && *ovar->sqlind)
    {
      Py_INCREF(Py_None);
      PyTuple_SetItem(result, i, Py_None);
    }
    else
    {
      /*----------------------------------------------------------------------*/
      /* For NUMERIC and DECIMAL columns we'll need to convert the integral   */
      /* data (Interbase internal) to floating point.                         */
      /*----------------------------------------------------------------------*/
      for (k = 0, divisor = 1; k > ovar->sqlscale; k--)
        divisor *= 10;

      /*----------------------------------------------------------------------*/
      /* Turn off the null bit to make our case statement cleaner.            */
      /*----------------------------------------------------------------------*/
      ctype = ovar->sqltype & ~1;
      switch (ctype)
      {
        case SQL_VARYING:
	  length = *((short *) ovar->sqldata);
	  ptr    = ovar->sqldata + sizeof(short);
	  PyTmp  = PyString_FromStringAndSize(ptr, length);
          break;

        case SQL_TEXT:
	  PyTmp  = PyString_FromStringAndSize(ovar->sqldata, ovar->sqllen);
          break;

        case SQL_SHORT:
	  if (ovar->sqlsubtype > 0)
	  {
	    short value;

	    value = *((short *) ovar->sqldata);
	    PyTmp = PyFloat_FromDouble((double) value / divisor);
          }
	  else
	  {
	    PyTmp = PyInt_FromLong(*((short *) ovar->sqldata));
          }
          break;

        case SQL_LONG:
	  if (ovar->sqlsubtype > 0)
	  {
	    long  value;

	    value = *((long *) ovar->sqldata);
	    PyTmp = PyFloat_FromDouble((double) value / divisor);
          }
	  else
	  {
	    PyTmp = PyLong_FromLong(*((long *) ovar->sqldata));
          }
          break;

        case SQL_FLOAT:
	  PyTmp = PyFloat_FromDouble(*((float *) ovar->sqldata));
          break;

        case SQL_DOUBLE:
        case SQL_D_FLOAT:
	  PyTmp = PyFloat_FromDouble(*((double *) ovar->sqldata));
          break;

	case SQL_INT64:
	  if (ovar->sqlsubtype > 0)
	  {
	    ISC_INT64 value;

	    value = *((ISC_INT64 *) ovar->sqldata);
	    PyTmp = PyFloat_FromDouble((double) value / divisor);
          }
	  else
	  {
	    PyTmp = PyLong_FromLongLong(*((ISC_INT64 *) ovar->sqldata));
          }
          break;

	case SQL_TYPE_TIME:
	{
	  PyObject *PyArgs;

	  isc_decode_sql_time((ISC_TIME *) ovar->sqldata, &dtime);

	  /*------------------------------------------------------------------*/
	  /*                                                                  */
	  /*                                                                  */
	  /*------------------------------------------------------------------*/
          PyArgs = Py_BuildValue("(iiiiiii)", 
			       ctype,
	                       1,
			       1,
			       1,
	                       dtime.tm_hour,
	                       dtime.tm_min,
	                       dtime.tm_sec);

          if (PyArgs == NULL)
	    return NULL;

          PyTmp = PyEval_CallObject(PyDateTime, PyArgs);
	  Py_XDECREF(PyArgs);

	  if (PyTmp == NULL)
	    return NULL;

	  break;
	}

        case SQL_TIMESTAMP:
	{
	  long     precision = ISC_TIME_SECONDS_PRECISION;
	  long     subseconds;
	  double   seconds;
	  ISC_TIME raw_time;
	  PyObject *PyArgs;

          isc_decode_timestamp((ISC_TIMESTAMP *) ovar->sqldata, &dtime);

	  raw_time   = ((ISC_TIMESTAMP *) ovar->sqldata)->timestamp_time;
	  subseconds = raw_time % precision;
	  seconds    = dtime.tm_sec + (double)((double)subseconds / precision);

          PyArgs = Py_BuildValue("(iiiiiif)", 
			       ctype,
	                       dtime.tm_year + 1900,
	                       dtime.tm_mon + 1,
	                       dtime.tm_mday,
	                       dtime.tm_hour,
	                       dtime.tm_min,
	                       seconds);

          if (PyArgs == NULL)
	    return NULL;

          PyTmp = PyEval_CallObject(PyDateTime, PyArgs);
	  Py_XDECREF(PyArgs);

	  if (PyTmp == NULL)
	    return NULL;

	  break;
        }

	case SQL_TYPE_DATE:
	{
	  PyObject *PyArgs;

	  isc_decode_sql_date((ISC_DATE *) ovar->sqldata, &dtime);

	  /*------------------------------------------------------------------*/
	  /*                                                                  */
	  /*                                                                  */
	  /*------------------------------------------------------------------*/
          PyArgs = Py_BuildValue("(iiiiiii)", 
			       ctype,
	                       dtime.tm_year + 1900,
	                       dtime.tm_mon + 1,
	                       dtime.tm_mday,
	                       dtime.tm_hour,
	                       dtime.tm_min,
	                       dtime.tm_sec);

          if (PyArgs == NULL)
	    return NULL;

          PyTmp = PyEval_CallObject(PyDateTime, PyArgs);
	  Py_XDECREF(PyArgs);

	  if (PyTmp == NULL)
	    return NULL;

	  break;
        }

	case SQL_BLOB:
	  PyTmp = getBlob(self, (ISC_QUAD *) ovar->sqldata);
	  if (PyTmp == NULL)
	  {
	    Py_DECREF(result);
	    return NULL;
	  }
	  break;

	case SQL_ARRAY:
	case SQL_QUAD:
        default:
          Py_INCREF(Py_None);
	  PyTmp = Py_None;
	  break;
      }

      if (PyTuple_SetItem(result, i, PyTmp))
      {
        Py_DECREF(result);
	return NULL;
      }
    }
  }

  return result;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorBindCols                                                           */
/*                                                                            */
/* This function binds the output columns in the XSQLVAR structures.  This    */
/* involves allocating the correct amount of storage based on the type and    */
/* size of the given column.                                                  */
/*                                                                            */
/******************************************************************************/
static long IBCursorBindCols (Cursor *self)
{
  short   i, n;
  short   ctype;
  short   size;
  XSQLDA  *sqlda;
  XSQLVAR *ovar;

  char emsg[128];

  PyObject *PyMsg;

  sqlda = self->osqlda;
  ovar  = self->osqlda->sqlvar;

  /*--------------------------------------------------------------------------*/
  /* Each XSQLVAR structure represents a column of the result set. For each   */
  /* structure turn off the null flag to simplify the case statement and then */
  /* allocate the correct amount of storage.                                  */
  /*--------------------------------------------------------------------------*/
  for (i = 0, n = sqlda->sqld; i < n; i++, ovar++)
  {
    ctype = ovar->sqltype & ~1;
    switch (ctype)
    {
      case SQL_TEXT:
        size = ovar->sqllen;
	break;

      case SQL_VARYING:
        size = ovar->sqllen + sizeof(short);
	break;

      case SQL_SHORT:
        size = sizeof(short);
	break;

      case SQL_LONG:
        size = sizeof(long);
	break;

      case SQL_FLOAT:
        size = sizeof(float);
	break;

      case SQL_DOUBLE:
      case SQL_D_FLOAT:
        size = sizeof(double);
	break;

      case SQL_TIMESTAMP:
        size = sizeof(ISC_TIMESTAMP);
	break;

      case SQL_TYPE_TIME:
        size = sizeof(ISC_TIME);
	break;

      case SQL_TYPE_DATE:
        size = sizeof(ISC_DATE);
	break;

      case SQL_INT64:
        size = sizeof(ISC_INT64);
	break;

      case SQL_BLOB:
      case SQL_ARRAY:
      case SQL_QUAD:
        size = sizeof(ISC_QUAD);
	break;

      default:
        PyMsg = PyDict_GetItemString(msgTable, "bindColumn");
	PyErr_SetObject(PyExc_StandardError, PyMsg);
        return -1;
    }

    /*------------------------------------------------------------------------*/
    /* Allocate storage for the data.                                         */
    /*------------------------------------------------------------------------*/
    Py_Free(ovar->sqldata);
    ovar->sqldata = (char *) Py_Malloc(size);
    if (!ovar->sqldata)
      return -1;

  }
  
  return 0;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorBindParm                                                           */
/*                                                                            */
/* This function binds one parameter to one XSQLVAR structure.  The steps     */
/* involve allocating storage for the data and then getting the value from    */
/* the python object and storing it in the data area.  The null indicator     */
/* is set to 'not null'.                                                      */
/*                                                                            */
/******************************************************************************/
static PyObject * 
IBCursorBindParm (Cursor *self, long index, PyObject *data)
{
  int i;
  long   long_value;
  short  ctype;

  XSQLDA   *sqlda;
  XSQLVAR  *var;
  PyObject *PyMsg;

  sqlda = self->isqlda;
  var = &sqlda->sqlvar[index];

  /*--------------------------------------------------------------------------*/
  /* Make sure we aren't indexing past the number of actual parameters.       */
  /*--------------------------------------------------------------------------*/
  if (index > sqlda->sqld)
  {
    setBadParmIndexException(index, sqlda->sqld);
    return NULL;
  }

  /*--------------------------------------------------------------------------*/
  /* Free up any storage previously allocated for the data.                   */
  /*--------------------------------------------------------------------------*/
  Py_Free(var->sqldata);

  /*--------------------------------------------------------------------------*/
  /* If None is passed then set the NULL flag and return;  Let IB choke if    */
  /* the column is defined as not null.  This way we'll get better messages.  */
  /*--------------------------------------------------------------------------*/
  if (data == Py_None)
  {
    *var->sqlind = -1;

    Py_INCREF(Py_None);
    return Py_None;
  }

  /*--------------------------------------------------------------------------*/
  /* Turn off the null bit to simplify the case statement.                    */
  /*--------------------------------------------------------------------------*/
  ctype = var->sqltype & ~1;
  switch (ctype)
  {
    case SQL_VARYING: 
      if (bindParmVarying(var, data, index))
        return NULL;

      break;

    case SQL_TEXT:
      if (bindParmText(var, data, index))
        return NULL;

      break;

    case SQL_SHORT:
      if (bindParmShort(var, data, index))
        return NULL;

      break;

    case SQL_LONG:
      if (bindParmLong(var, data, index))
        return NULL;

      break;

    case SQL_FLOAT:
      if (bindParmFloat(var, data, index))
        return NULL;

      break;

    case SQL_DOUBLE:
    case SQL_D_FLOAT:
      if (bindParmDouble(var, data, index))
        return NULL;

      break;

    case SQL_TIMESTAMP:
      if (bindParmTimestamp(var, data, index))
        return NULL;

      break;

    case SQL_TYPE_TIME:
      if (bindParmTime(var, data, index))
        return NULL;

      break;

    case SQL_TYPE_DATE:
      if (bindParmDate(var, data, index))
        return NULL;

      break;

    case SQL_INT64:
      if (bindParmInt64(var, data, index))
        return NULL;

      break;

    case SQL_BLOB:
      if (bindParmBlob(var, data, index, self))
        return NULL;

      break;
      
    case SQL_ARRAY:
    case SQL_QUAD:
    default:
      return NULL;
  }

  *var->sqlind = 0;

  Py_INCREF(Py_None);
  return Py_None;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorReallocXDA                                                         */
/*                                                                            */
/* This function reallocates the input or output XSQLDA data structure.       */
/*                                                                            */
/******************************************************************************/
static long IBCursorReallocXDA (Cursor *self, char *type)
{
  short i, n;
  char emsg[128];
  XSQLVAR *var;

  PyObject *PyMsg;

  (void) IBCursorFreeCols(self, type);

  if (strcmp(type, "isqlda") == 0)
  {
    n = self->isqlda->sqld;
    self->isqlda = (XSQLDA *) Py_Realloc(self->isqlda, XSQLDA_LENGTH(n));
    if (!self->isqlda)
      return -1;

    self->isqlda->sqln    = self->isqlda->sqld;
    self->isqlda->version = SQLDA_VERSION1;

    /*------------------------------------------------------------------------*/
    /* Set sqldata to NULL.  Found by Brad Clements.  07/31/2000              */
    /*------------------------------------------------------------------------*/
    for (i = 0, var = self->isqlda->sqlvar; i < n; i++, var++)
    {
      var->sqlind = (short *) Py_Malloc(sizeof(short));
      if (!var->sqlind)
        return -1;

      var->sqldata = NULL;
    }
  }
  else if (strcmp(type, "osqlda") == 0)
  {
    n = self->osqlda->sqld;
    self->osqlda = (XSQLDA *) Py_Realloc(self->osqlda, XSQLDA_LENGTH(n));
    if (!self->osqlda)
      return -1;

    self->osqlda->sqln    = self->osqlda->sqld;
    self->osqlda->version = SQLDA_VERSION1;

    /*------------------------------------------------------------------------*/
    /* Set sqldata to NULL.  Found by Brad Clements.  07/31/2000              */
    /*------------------------------------------------------------------------*/
    for (i = 0, var = self->osqlda->sqlvar; i < n; i++, var++)
    {
      var->sqlind = (short *) Py_Malloc(sizeof(short));
      if (!var->sqlind)
        return -1;

      var->sqldata = NULL;
    }
  }
  else
  {
    PyMsg = PyDict_GetItemString(msgTable, "badXSQLDA");
    PyErr_SetObject(PyExc_StandardError, PyMsg);
    return -1;
  }

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorSetAttr                                                            */
/*                                                                            */
/* This function sets data in the C Cursor object as passed from python.      */
/*                                                                            */
/******************************************************************************/
static int
IBCursorSetAttr (Cursor *self, char *name, PyObject *data)
{
  if (strcmp(name, "stmt_handle") == 0)
  {
    self->stmt_handle = (isc_stmt_handle) PyLong_AsLong(data);
  }
  
  return 0;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorGetAttr                                                            */
/*                                                                            */
/* This function returns data from the cursor object to python in a form that */
/* can then be passed from python to an Interbase API function.               */
/*                                                                            */
/******************************************************************************/
static PyObject *
IBCursorGetAttr (Cursor *self, char *name)
{
  long tmp_long;
  char tmp_string[256];

  if (strcmp(name, "status") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) self->status, "_ISC_STATUS_p");
    return PyString_FromString(tmp_string);
  }

  if (strcmp(name, "stmt_handle") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) &self->stmt_handle, "_isc_stmt_handle_p");
    return PyString_FromString(tmp_string);
  }

  if (strcmp(name, "db_handle") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) &self->connect->db_handle, 
                 "_isc_db_handle_p");
    return PyString_FromString(tmp_string);
  }

  if (strcmp(name, "isqlda") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) self->isqlda, "_XSQLDA_p");
    return PyString_FromString(tmp_string);
  }

  if (strcmp(name, "osqlda") == 0)
  {
    SWIG_MakePtr(tmp_string, (void *) self->osqlda, "_XSQLDA_p");
    return PyString_FromString(tmp_string);
  }

  if (strcmp(name, "isqln") == 0)
  {
    tmp_long = (long) self->isqlda->sqln;
    return PyLong_FromLong(tmp_long);
  }

  if (strcmp(name, "isqld") == 0)
  {
    tmp_long = (long) self->isqlda->sqld;
    return PyLong_FromLong(tmp_long);
  }

  if (strcmp(name, "osqln") == 0)
  {
    tmp_long = (long) self->osqlda->sqln;
    return PyLong_FromLong(tmp_long);
  }

  if (strcmp(name, "osqld") == 0)
  {
    tmp_long = (long) self->osqlda->sqld;
    return PyLong_FromLong(tmp_long);
  }

  return Py_None;
}


/******************************************************************************/
/*                                                                            */
/* IBCursor                                                                   */
/*                                                                            */
/* This is the constructor for the C interface Cursor object.                 */
/*                                                                            */
/******************************************************************************/
static PyObject *
IBCursor (Connection *connect, long columns)
{
  int      i;
  long     size;
  Cursor  *self;
  XSQLVAR *var;

  if ((self = PyObject_NEW(Cursor, &CursorType)) == NULL)
    return NULL;

  self->stmt_handle = 0L;
  self->connect     = connect;

  size = XSQLDA_LENGTH(columns);

  if ((self->isqlda = (XSQLDA *) Py_Malloc(size)) == NULL)
    return NULL;

  if ((self->osqlda = (XSQLDA *) Py_Malloc(size)) == NULL)
    return NULL;

  memset((char *) self->isqlda, 0, size);
  memset((char *) self->osqlda, 0, size);

  self->isqlda->version = self->osqlda->version = SQLDA_VERSION1;
  self->isqlda->sqln    = self->osqlda->sqln    = (short) columns;

  /*--------------------------------------------------------------------------*/
  /* Allocate storage for the input indicator variables because this storage  */
  /* is unchanging in size.                                                   */
  /*--------------------------------------------------------------------------*/
  for (i = 0, var = self->isqlda->sqlvar; i < columns; i++, var++)
  {
    var->sqlind = (short *) Py_Malloc(sizeof(short));
    if (!var->sqlind)
      return NULL;
  }

  /*--------------------------------------------------------------------------*/
  /* Allocate storage for the output indicator variables because this storage */
  /* is unchanging in size.                                                   */
  /*--------------------------------------------------------------------------*/
  for (i = 0, var = self->osqlda->sqlvar; i < columns; i++, var++)
  {
    var->sqlind = (short *) Py_Malloc(sizeof(short));
    if (!var->sqlind)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS

  isc_dsql_allocate_statement(self->status, 
                              &self->connect->db_handle,
			      &self->stmt_handle);

  Py_END_ALLOW_THREADS

  if (DB_ERROR(self->status))
  {
    (void) IBSetException(self->status);
    Py_DECREF(self);

    return NULL;
  }

  return (PyObject *) self;
}


/******************************************************************************/
/*                                                                            */
/* IBCursorDeallocate                                                         */
/*                                                                            */
/* This is the destructor for the C interface Cursor object.                  */
/*                                                                            */
/******************************************************************************/
static void
IBCursorDeallocate (Cursor *self)
{
  short i, n;
  long sv[20];
  XSQLVAR *var;

  if (self)
  {
    if (self->stmt_handle)
    {
      Py_BEGIN_ALLOW_THREADS

      isc_dsql_free_statement(sv, &self->stmt_handle, DSQL_drop);

      Py_END_ALLOW_THREADS
    }

    /*------------------------------------------------------------------------*/
    /* For the input SQLDA free the memory allocated for the SQL indicator    */
    /* variables and then loop and free the memory for the SQL data.          */
    /*------------------------------------------------------------------------*/
    if (self->isqlda)
    {
      var = self->isqlda->sqlvar;

      for (i = 0, n = self->isqlda->sqln; i < n; i++, var++)
      {
        Py_Free(var->sqldata);
        Py_Free(var->sqlind);
      }

      Py_Free(self->isqlda);
    }

    /*------------------------------------------------------------------------*/
    /* For the output SQLDA free the memory allocated for the SQL indicator   */
    /* variables and then loop and free the memory for the SQL data.          */
    /*------------------------------------------------------------------------*/
    if (self->osqlda)
    {
      var = self->osqlda->sqlvar;

      for (i = 0, n = self->osqlda->sqln; i < n; i++, var++)
      {
        Py_Free(var->sqldata);
        Py_Free(var->sqlind);
      }

      Py_Free(self->osqlda);
    }
  }

  PyMem_DEL(self);
}

%}


%{  /* Start of SWIG-Free Section */

/******************************************************************************/
/*                                                                            */
/* getBlob                                                                    */
/*                                                                            */
/* This function is used to fetch a blob object given the blob ID.  The blob  */
/* is returned to the caller as a Rpython Buffer object.                      */
/*                                                                            */
/******************************************************************************/
static PyObject *
getBlob (Cursor *self, ISC_QUAD *blob_id)
{
  char *buffer;

  unsigned short segment_size;
  unsigned short actual_length;

  ISC_STATUS rc; 
  ISC_STATUS tmp_status[20];
  isc_blob_handle blob_handle = NULL;

  PyObject *PyTmp = NULL;
  PyObject *PyStr = NULL;

  segment_size = self->connect->page_size;

  Py_BEGIN_ALLOW_THREADS

  /*--------------------------------------------------------------------------*/
  /* We retrieve a BLOB by opening it and then entering a loop and fetch one  */
  /* segment at a time until all segments have been read.                     */
  /*--------------------------------------------------------------------------*/
  isc_open_blob2(self->status,
                 &self->connect->db_handle,
                 &self->connect->tr_handle,
		 &blob_handle,
		 blob_id,
		 0,
		 NULL);


  if (!DB_ERROR(self->status))
  {
    /*------------------------------------------------------------------------*/
    /* The segment size is set equal to the database page size.  This will    */
    /* help to maximize throughput.                                           */
    /*------------------------------------------------------------------------*/
    buffer = (char *) Py_Malloc(segment_size);
    if (!buffer)
      return NULL;

    PyStr = PyString_FromString("");

    do
    {
      rc = isc_get_segment(self->status,
                           &blob_handle,
			   &actual_length,
			   segment_size,
			   buffer);

      /*----------------------------------------------------------------------*/
      /* A return code of 0 means a segment was read and more segments may    */
      /* exist.  A return code of isc_segment means that the segment was      */
      /* larger than the buffer so more data needs to be read.  A return      */
      /* code if isc_segstr_eof means we reached the end and we are done.     */
      /*----------------------------------------------------------------------*/
      if (rc == 0 || rc == isc_segment)
      {
        PyTmp = PyString_FromStringAndSize(buffer, actual_length);
        PyString_ConcatAndDel(&PyStr, PyTmp);
        if (PyStr == NULL)
          break;
      }
    }
    while (rc != isc_segstr_eof);

    isc_close_blob(tmp_status, &blob_handle);

    Py_Free(buffer);
  }

  Py_END_ALLOW_THREADS

  /*--------------------------------------------------------------------------*/
  /* If true the isc_get_segment had a problem.                               */
  /*--------------------------------------------------------------------------*/
  if (rc != isc_segstr_eof)
  {
    (void) IBSetException(self->status);
    Py_XDECREF(PyStr);
    return NULL;
  }

  /*--------------------------------------------------------------------------*/
  /* If true the isc_close_blob had a problem.                                */
  /*--------------------------------------------------------------------------*/
  if (DB_ERROR(tmp_status))
  {
    (void) IBSetException(tmp_status);
    Py_XDECREF(PyStr);
    return NULL;
  }

  if (PyStr == NULL)
    return NULL;

  return PyBuffer_FromObject(PyStr, 0, Py_END_OF_BUFFER);
}


/******************************************************************************/
/*                                                                            */
/* putBlob                                                                    */
/*                                                                            */
/* This function creates a blob and returns the blob ID to the caller.        */
/*                                                                            */
/******************************************************************************/
static short
putBlob (Cursor *self, ISC_QUAD *blob_id, PyObject *blob)
{
  long n;
  char *buffer;
  unsigned long  length;
  unsigned short segment_size;

  ISC_STATUS rc; 
  ISC_STATUS tmp_status[20];
  isc_blob_handle blob_handle = NULL;

  PyObject *PyStr = NULL;

  /*--------------------------------------------------------------------------*/
  /*                                                                          */
  /*--------------------------------------------------------------------------*/
  if (PyString_Check(blob))
  {
    buffer = PyString_AsString(blob);
    length = PyString_Size(blob);
  }
  else if (PyBuffer_Check(blob))
  {
    PyStr  = PyObject_Str(blob);
    buffer = PyString_AsString(PyStr);
    length = PyString_Size(PyStr);
  }
  else
  {
    return -1;
  }

  Py_BEGIN_ALLOW_THREADS

  /*--------------------------------------------------------------------------*/
  /* We create or update a BLOB by creating a new blob, writing the blob data */
  /* and then by returning the blob id to the caller which gets stored in     */
  /* the actual database column as a QUAD value.                              */
  /*--------------------------------------------------------------------------*/
  isc_create_blob2(self->status,
                 &self->connect->db_handle,
                 &self->connect->tr_handle,
		 &blob_handle,
		 blob_id,
		 0,
		 NULL);

  if (!DB_ERROR(self->status))
  {
    /*------------------------------------------------------------------------*/
    /* We can only write a maximum of USHRT_MAX at a time so determine if     */
    /* the length of data remaining to be written is smaller or greater than  */
    /* USHRT_MAX and then write the minimum size.  Do this until we are done. */
    /*------------------------------------------------------------------------*/
    for (n = rc = 0; (n < length) && (rc == 0); n += segment_size)
    {
      segment_size = MIN(USHRT_MAX, length - n);
      rc = isc_put_segment(self->status,
                           &blob_handle,
			   segment_size,
			   buffer + n);

    }

    isc_close_blob(tmp_status, &blob_handle);
  }

  Py_END_ALLOW_THREADS

  Py_XDECREF(PyStr);

  /*--------------------------------------------------------------------------*/
  /* If true the create or put had a problem.                                 */
  /*--------------------------------------------------------------------------*/
  if (DB_ERROR(self->status))
  {
    (void) IBSetException(self->status);
    return -1;
  }

  /*--------------------------------------------------------------------------*/
  /* If true the close had a problem.                                         */
  /*--------------------------------------------------------------------------*/
  if (DB_ERROR(tmp_status))
  {
    (void) IBSetException(tmp_status);
    return -1;
  }

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* getPrecision                                                               */
/*                                                                            */
/* This function makes a call to the system tables to get the precision for   */
/* a given column and table.                                                  */
/*                                                                            */
/******************************************************************************/
static short 
getPrecision (Cursor *self, char *col_name)
{
  struct
  {
    short null;
    short data;
  }
  result;

  XSQLDA  *sqlda;
  XSQLVAR *var;

  char  table[64];
  char  column[64];
  static short not_null = 0;

  static XSQLDA *i_sqlda = NULL;
  static XSQLDA *o_sqlda = NULL;

  char *sql_stmt = "SELECT A.RDB$FIELD_PRECISION "
		     "FROM RDB$FIELDS A, "
			  "RDB$RELATION_FIELDS B "
                    "WHERE A.RDB$FIELD_NAME = B.RDB$FIELD_SOURCE "
		      "AND B.RDB$FIELD_NAME = ? "
		      "AND B.RDB$RELATION_NAME = ?";

  sqlda = self->osqlda;
  var   = self->osqlda->sqlvar;

  /*--------------------------------------------------------------------------*/
  /* Copy the table and column names into the appropriate buffers.            */
  /*--------------------------------------------------------------------------*/
  memset(table, 0, sizeof(table));
  memcpy(table, var->relname, var->relname_length);
  strcpy(column, col_name);

  /*--------------------------------------------------------------------------*/
  /* The DA's can be shared among threads so allocate once for better         */
  /* performance.  Here we allocate and initialize the input DA.              */
  /*--------------------------------------------------------------------------*/
  if (i_sqlda == NULL)
  {
    short size = XSQLDA_LENGTH(2);

    if ((i_sqlda = (XSQLDA *) Py_Malloc(size)) == NULL)
      return -1;

    memset((char *) i_sqlda, 0, size);
    i_sqlda->version = SQLDA_VERSION1;
    i_sqlda->sqln    = 2;
    i_sqlda->sqld    = 2;

    i_sqlda->sqlvar[0].sqltype = SQL_TEXT + 1;
    i_sqlda->sqlvar[0].sqlind  = &not_null;

    i_sqlda->sqlvar[1].sqltype = SQL_TEXT + 1;
    i_sqlda->sqlvar[1].sqlind  = &not_null;
  }

  /*--------------------------------------------------------------------------*/
  /* Allocate and initialize the output DA.                                   */
  /*--------------------------------------------------------------------------*/
  if (o_sqlda == NULL)
  {
    short size = XSQLDA_LENGTH(1);

    if ((o_sqlda = (XSQLDA *) Py_Malloc(size)) == NULL)
      return -1;

    memset((char *) o_sqlda, 0, size);
    o_sqlda->version = SQLDA_VERSION1;
    o_sqlda->sqln    = 1;
    o_sqlda->sqld    = 1;

    o_sqlda->sqlvar[0].sqltype = SQL_SHORT + 1;
  }

  /*--------------------------------------------------------------------------*/
  /* Set some of the SQLVAR data members.  Since column and table are on the  */
  /* stack we should be thread safe.                                          */
  /*--------------------------------------------------------------------------*/
  i_sqlda->sqlvar[0].sqldata = column;
  i_sqlda->sqlvar[0].sqllen  = strlen(column);

  i_sqlda->sqlvar[1].sqldata = table;
  i_sqlda->sqlvar[1].sqllen  = var->relname_length;

  o_sqlda->sqlvar[0].sqldata = (char *) &result.data;
  o_sqlda->sqlvar[0].sqllen  = sizeof(short);
  o_sqlda->sqlvar[0].sqlind  = &result.null;

  Py_BEGIN_ALLOW_THREADS

  isc_dsql_exec_immed2(self->status,
		       &self->connect->db_handle,
		       &self->connect->tr_handle,
		       0,
		       sql_stmt,
		       SQL_DIALECT_CURRENT,
		       i_sqlda,
		       o_sqlda);

  Py_END_ALLOW_THREADS

  if (DB_ERROR(self->status))
  {
    (void) IBSetException(self->status);

    return -1;
  }

  if (result.null < 0)
    return 0;

  return result.data;
}


%}  /* End of SWIG-Free Section */

